from types import NoneType
from typing import List, Optional, Set, Tuple, Type, Union

import pytest

from jmux.awaitable import (
    AwaitableValue,
    StreamableValues,
    UnderlyingGenericMixin,
)
from jmux.demux import JMux
from jmux.error import ParsePrimitiveError
from jmux.helpers import (
    deconstruct_flat_type,
    extract_types_from_generic_alias,
    get_main_type,
    str_to_bool,
)


@pytest.mark.parametrize(
    "TargetType,expected_tuple",
    [
        (int, {int}),
        (Union[int], {int}),
        #
        (Optional[int], {int, NoneType}),
        (int | None, {int, NoneType}),
        #
        (Union[int, None], {int, NoneType}),
        (int | NoneType, {int, NoneType}),
        (Union[int, NoneType], {int, NoneType}),
        #
        (int | str, {str, int}),
        (Union[int, str], {str, int}),
        #
        (int | str | NoneType, {str, int, NoneType}),
        (Union[int, str, NoneType], {str, int, NoneType}),
        #
        (JMux, {JMux}),
        (Union[JMux], {JMux}),
        #
        (JMux | None, {JMux, NoneType}),
        (Union[JMux, None], {JMux, NoneType}),
        (Union[JMux, NoneType], {JMux, NoneType}),
        #
        (JMux | NoneType, {JMux, NoneType}),
        (Union[JMux, NoneType], {JMux, NoneType}),
    ],
)
def test_deconstruct_flat_types(
    TargetType: Type[UnderlyingGenericMixin], expected_tuple: Tuple[Type, Set[Type]]
):
    underlying_types = deconstruct_flat_type(TargetType)

    assert underlying_types == expected_tuple


class NestedObject(JMux):
    key: AwaitableValue[str]


# fmt: off
@pytest.mark.parametrize(
    "TargetType,expected_tuple",
    [
        (int, ({int}, set())),
        (str, ({str}, set())),
        (Optional[int], ({int, NoneType}, set())),
        (int | None, ({int, NoneType}, set())),
        (List[int], ({list}, {int})),
        (list[int], ({list}, {int})),
        (List[int | None], ({list}, {int, NoneType})),
        (list[int | None], ({list}, {int, NoneType})),
        (List[int] | None, ({list, NoneType}, {int})),
        (list[int] | None, ({list, NoneType}, {int})),
        (Optional[List[int]], ({list, NoneType}, {int})),
        (Optional[list[int]], ({list, NoneType}, {int})),
        (AwaitableValue[int], ({AwaitableValue}, {int})),
        (AwaitableValue[float], ({AwaitableValue}, {float})),
        (AwaitableValue[str], ({AwaitableValue}, {str})),
        (AwaitableValue[bool], ({AwaitableValue}, {bool})),
        (AwaitableValue[NestedObject], ({AwaitableValue}, {NestedObject})),
        (AwaitableValue[int | None], ({AwaitableValue}, {int, NoneType})),
        (AwaitableValue[float | None], ({AwaitableValue}, {float, NoneType})),
        (AwaitableValue[str | None], ({AwaitableValue}, {str, NoneType})),
        (AwaitableValue[bool | None], ({AwaitableValue}, {bool, NoneType})),
        (AwaitableValue[NestedObject | None], ({AwaitableValue}, {NestedObject, NoneType})),
        (StreamableValues[int], ({StreamableValues}, {int})),
        (StreamableValues[float], ({StreamableValues}, {float})),
        (StreamableValues[str], ({StreamableValues}, {str})),
        (StreamableValues[bool], ({StreamableValues}, {bool})),
        (StreamableValues[NestedObject], ({StreamableValues}, {NestedObject})),
    ],
)
# fmt: on
def test_extract_types_from_generic_alias(
    TargetType: Type[UnderlyingGenericMixin], expected_tuple: Tuple[Type, Set[Type]]
):
    underlying_types = extract_types_from_generic_alias(TargetType)

    assert underlying_types == expected_tuple


def test_str_to_bool__true():
    assert str_to_bool("true") is True


def test_str_to_bool__false():
    assert str_to_bool("false") is False


@pytest.mark.parametrize(
    "invalid_input",
    [
        "True",
        "False",
        "TRUE",
        "FALSE",
        "1",
        "0",
        "yes",
        "no",
        "",
        " ",
        "  true",
        "true  ",
        " true ",
        "tRue",
        "fAlse",
        "null",
        "None",
        "t",
        "f",
    ],
)
def test_str_to_bool__invalid_input_raises_parse_primitive_error(invalid_input: str):
    with pytest.raises(ParsePrimitiveError):
        str_to_bool(invalid_input)


def test_get_main_type__single_type():
    assert get_main_type({int}) is int
    assert get_main_type({str}) is str
    assert get_main_type({float}) is float
    assert get_main_type({bool}) is bool
    assert get_main_type({NoneType}) is NoneType
    assert get_main_type({JMux}) is JMux


def test_get_main_type__type_with_none():
    assert get_main_type({int, NoneType}) is int
    assert get_main_type({str, NoneType}) is str
    assert get_main_type({float, NoneType}) is float
    assert get_main_type({bool, NoneType}) is bool
    assert get_main_type({JMux, NoneType}) is JMux


def test_get_main_type__empty_set_raises_type_error():
    with pytest.raises(TypeError):
        get_main_type(set())


def test_get_main_type__multiple_non_none_types_raises_type_error():
    with pytest.raises(TypeError):
        get_main_type({int, str})


def test_get_main_type__multiple_types_with_none_raises_type_error():
    with pytest.raises(TypeError):
        get_main_type({int, str, NoneType})


def test_get_main_type__does_not_modify_input():
    input_set = {int, NoneType}
    original_copy = input_set.copy()
    get_main_type(input_set)
    assert input_set == original_copy


@pytest.mark.parametrize(
    "input_type",
    [
        NestedObject,
        AwaitableValue,
        StreamableValues,
    ],
)
def test_get_main_type__custom_types(input_type: Type):
    assert get_main_type({input_type}) == input_type
    assert get_main_type({input_type, NoneType}) == input_type


def test_deconstruct_flat_type__raises_on_list():
    with pytest.raises(TypeError):
        deconstruct_flat_type(List[int])


def test_deconstruct_flat_type__raises_on_generic_class():
    with pytest.raises(TypeError):
        deconstruct_flat_type(AwaitableValue[int])


def test_deconstruct_flat_type__none_type_directly():
    assert deconstruct_flat_type(None) == {NoneType}
    assert deconstruct_flat_type(NoneType) == {NoneType}


def test_extract_types_from_generic_alias__raises_on_dict():
    with pytest.raises(TypeError):
        extract_types_from_generic_alias(dict[str, int])


def test_extract_types_from_generic_alias__raises_on_tuple():
    with pytest.raises(TypeError):
        extract_types_from_generic_alias(tuple[int, str])


def test_extract_types_from_generic_alias__raises_on_multi_type_union_in_generic():
    with pytest.raises(TypeError):
        extract_types_from_generic_alias(AwaitableValue[int | str])


def test_extract_types_from_generic_alias__raises_on_union_without_none_in_generic():
    with pytest.raises(TypeError):
        extract_types_from_generic_alias(AwaitableValue[int | float])



from types import NoneType
from typing import Set, Tuple, Type, get_args, get_origin

from jmux.error import ParsePrimitiveError
from jmux.types import TYPES_LIKE_NONE, TYPES_LIKE_UNION


def str_to_bool(s: str) -> bool:
    if s == "true":
        return True
    elif s == "false":
        return False
    else:
        raise ParsePrimitiveError(
            f"Cannot convert string '{s}' to boolean. Expected 'true' or 'false', got"
            f" '{s}'."
        )


def extract_types_from_generic_alias(UnknownType: Type) -> Tuple[Set[Type], Set[Type]]:
    Origin: Type | None = get_origin(UnknownType)
    if Origin is None:
        return {UnknownType}, set()
    if Origin in TYPES_LIKE_UNION:
        deconstructed = deconstruct_flat_type(UnknownType)
        maybe_list_types = [
            subtypes for subtypes in deconstructed if get_origin(subtypes) is list
        ]
        if len(maybe_list_types) == 1:
            list_based_type = maybe_list_types[0]
            non_list_types = deconstructed - {list_based_type}
            main_type, subtype = extract_types_from_generic_alias(list_based_type)
            return non_list_types | main_type, subtype
        return deconstructed, set()

    type_args = get_args(UnknownType)
    if len(type_args) != 1:
        raise TypeError(
            f"Only single type generics can be deconstruct with this function, "
            f"got {type_args}."
        )

    Generic: Type = type_args[0]
    type_set = deconstruct_flat_type(Generic)
    if len(type_set) == 1:
        return {Origin}, type_set
    if len(type_set) != 2:
        raise TypeError(
            f"Union type must have exactly two types in its union, "
            f"got {get_args(Generic)}."
        )
    if NoneType not in get_args(Generic):
        raise TypeError(
            "Union type must include NoneType if it is used as a generic argument."
        )
    return {Origin}, type_set


def deconstruct_flat_type(UnknownType: Type) -> Set[Type]:
    Origin: Type | None = get_origin(UnknownType)
    if UnknownType in TYPES_LIKE_NONE:
        return {NoneType}
    if Origin is None:
        return {UnknownType}
    if Origin in TYPES_LIKE_UNION:
        type_args = get_args(UnknownType)
        return set(type_args)
    raise TypeError(
        f"Unknown type {UnknownType} is not a Union or optional type, "
        "only only those types and their syntactic sugar are supported "
        "for flat deconstruction."
    )


def get_main_type(type_set: Set[Type]) -> Type:
    type_set_copy: Set[Type] = type_set.copy()
    if NoneType in type_set_copy and len(type_set_copy) == 2:
        type_set_copy.remove(NoneType)
    if len(type_set_copy) != 1:
        raise TypeError(
            f"Expected exactly one type, got {type_set_copy}. If you want to allow "
            "NoneType, use Union[int, NoneType]."
        )
    return type_set_copy.pop()



import pytest

from jmux.pda import PushDownAutomata
from jmux.types import Mode, State


@pytest.mark.parametrize(
    "start_state",
    [
        State.START,
        State.END,
        State.ERROR,
        State.EXPECT_KEY,
        State.PARSING_STRING,
    ],
)
def test_pda_initial_state(start_state: State):
    pda = PushDownAutomata[Mode, State](start_state)
    assert pda.state == start_state


@pytest.mark.parametrize(
    "start_state",
    [
        State.START,
        State.END,
    ],
)
def test_pda_initial_stack_is_empty(start_state: State):
    pda = PushDownAutomata[Mode, State](start_state)
    assert pda.stack == []
    assert pda.top is None


@pytest.mark.parametrize(
    "new_state",
    [
        State.EXPECT_KEY,
        State.EXPECT_COLON,
        State.EXPECT_VALUE,
        State.PARSING_STRING,
        State.END,
    ],
)
def test_pda_set_state(new_state: State):
    pda = PushDownAutomata[Mode, State](State.START)
    pda.set_state(new_state)
    assert pda.state == new_state


def test_pda_set_state_multiple_times():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.set_state(State.EXPECT_KEY)
    assert pda.state == State.EXPECT_KEY
    pda.set_state(State.PARSING_STRING)
    assert pda.state == State.PARSING_STRING
    pda.set_state(State.END)
    assert pda.state == State.END


@pytest.mark.parametrize(
    "mode",
    [
        Mode.ROOT,
        Mode.OBJECT,
        Mode.ARRAY,
    ],
)
def test_pda_push_single(mode: Mode):
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(mode)
    assert pda.stack == [mode]
    assert pda.top == mode


def test_pda_push_multiple():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.push(Mode.OBJECT)
    pda.push(Mode.ARRAY)
    assert pda.stack == [Mode.ROOT, Mode.OBJECT, Mode.ARRAY]
    assert pda.top == Mode.ARRAY


def test_pda_pop_single():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    result = pda.pop()
    assert result == Mode.ROOT
    assert pda.stack == []
    assert pda.top is None


def test_pda_pop_multiple():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.push(Mode.OBJECT)
    pda.push(Mode.ARRAY)

    result1 = pda.pop()
    assert result1 == Mode.ARRAY
    assert pda.top == Mode.OBJECT

    result2 = pda.pop()
    assert result2 == Mode.OBJECT
    assert pda.top == Mode.ROOT

    result3 = pda.pop()
    assert result3 == Mode.ROOT
    assert pda.top is None


def test_pda_pop_empty_stack_raises_index_error():
    pda = PushDownAutomata[Mode, State](State.START)
    with pytest.raises(IndexError, match="PDA stack is empty"):
        pda.pop()


def test_pda_pop_after_emptying_stack_raises_index_error():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.pop()
    with pytest.raises(IndexError, match="PDA stack is empty"):
        pda.pop()


def test_pda_push_pop_interleaved():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.push(Mode.OBJECT)
    assert pda.pop() == Mode.OBJECT

    pda.push(Mode.ARRAY)
    pda.push(Mode.OBJECT)
    assert pda.stack == [Mode.ROOT, Mode.ARRAY, Mode.OBJECT]
    assert pda.top == Mode.OBJECT

    assert pda.pop() == Mode.OBJECT
    assert pda.pop() == Mode.ARRAY
    assert pda.top == Mode.ROOT


def test_pda_top_does_not_modify_stack():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.push(Mode.OBJECT)

    _ = pda.top
    _ = pda.top
    _ = pda.top

    assert pda.stack == [Mode.ROOT, Mode.OBJECT]


def test_pda_stack_property_returns_internal_list():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    stack = pda.stack
    assert stack is pda._stack


def test_pda_with_string_types():
    pda = PushDownAutomata[str, str]("initial")
    assert pda.state == "initial"
    pda.set_state("next")
    assert pda.state == "next"

    pda.push("context1")
    pda.push("context2")
    assert pda.top == "context2"
    assert pda.pop() == "context2"


def test_pda_with_int_types():
    pda = PushDownAutomata[int, int](0)
    assert pda.state == 0
    pda.set_state(1)
    assert pda.state == 1

    pda.push(100)
    pda.push(200)
    assert pda.top == 200
    assert pda.stack == [100, 200]


def test_pda_many_push_operations():
    pda = PushDownAutomata[Mode, State](State.START)
    for _ in range(100):
        pda.push(Mode.OBJECT)
    assert len(pda.stack) == 100
    assert pda.top == Mode.OBJECT


def test_pda_many_pop_operations():
    pda = PushDownAutomata[int, int](0)
    for i in range(100):
        pda.push(i)

    for i in range(99, -1, -1):
        assert pda.pop() == i

    assert pda.stack == []
    assert pda.top is None


def test_pda_top_returns_last_pushed_incrementally():
    pda = PushDownAutomata[Mode, State](State.START)
    assert pda.top is None

    pda.push(Mode.ROOT)
    assert pda.top == Mode.ROOT

    pda.push(Mode.OBJECT)
    assert pda.top == Mode.OBJECT

    pda.push(Mode.ARRAY)
    assert pda.top == Mode.ARRAY


def test_pda_state_is_readable_property():
    pda = PushDownAutomata[Mode, State](State.START)
    assert pda.state == State.START

    pda.set_state(State.EXPECT_KEY)
    state_value = pda.state
    assert state_value == State.EXPECT_KEY
    assert isinstance(state_value, State)


def test_pda_typical_json_parsing_flow():
    pda = PushDownAutomata[Mode, State](State.START)

    pda.set_state(State.EXPECT_KEY)
    pda.push(Mode.ROOT)
    assert pda.state == State.EXPECT_KEY
    assert pda.top == Mode.ROOT

    pda.set_state(State.PARSING_KEY)
    assert pda.state == State.PARSING_KEY

    pda.set_state(State.EXPECT_COLON)
    pda.set_state(State.EXPECT_VALUE)
    pda.set_state(State.PARSING_STRING)

    pda.push(Mode.ARRAY)
    assert pda.stack == [Mode.ROOT, Mode.ARRAY]

    pda.set_state(State.EXPECT_COMMA_OR_EOC)
    pda.pop()
    assert pda.top == Mode.ROOT

    pda.set_state(State.END)
    pda.pop()
    assert pda.stack == []
    assert pda.state == State.END


def test_pda_stack_returns_full_contents():
    pda = PushDownAutomata[Mode, State](State.START)
    pda.push(Mode.ROOT)
    pda.push(Mode.OBJECT)
    pda.push(Mode.ARRAY)

    stack_contents = pda.stack
    assert stack_contents == [Mode.ROOT, Mode.OBJECT, Mode.ARRAY]
    assert len(stack_contents) == 3



from typing import List

import pytest

from jmux.decoder import StringEscapeDecoder


# fmt: off
@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("foo bar", "foo bar"),
        ("foo\"bar", 'foo"bar'),
        ("foo bar\"", 'foo bar"'),
        ("foo\\\\bar", "foo\\bar"),
        ("foo\\bbar", "foo\bbar"),
        ("foo\\tbar", "foo\tbar"),
        ("foo\\rbar", "foo\rbar"),
        ("foo\\fbar", "foo\fbar"),
        ("foo\\/bar", "foo/bar"),
        ("foo\\/bar", "foo/bar"),
    ],
)
# fmt: on
@pytest.mark.anyio
async def test_string_decoder__parameterized(stream: str, expected_string: List[str]):
    decoder = StringEscapeDecoder()

    for ch in stream:
        decoder.push(ch)

    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\n", "\n"),
        ("hello\\nworld", "hello\nworld"),
        ("line1\\nline2\\nline3", "line1\nline2\nline3"),
        ("\\n\\n\\n", "\n\n\n"),
    ],
)
def test_string_decoder__newline_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\t", "\t"),
        ("col1\\tcol2\\tcol3", "col1\tcol2\tcol3"),
        ("\\t\\t\\t", "\t\t\t"),
    ],
)
def test_string_decoder__tab_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\r", "\r"),
        ("line1\\rline2", "line1\rline2"),
        ("\\r\\n", "\r\n"),
    ],
)
def test_string_decoder__carriage_return_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\b", "\b"),
        ("hello\\bworld", "hello\bworld"),
    ],
)
def test_string_decoder__backspace_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\f", "\f"),
        ("page1\\fpage2", "page1\fpage2"),
    ],
)
def test_string_decoder__formfeed_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\/", "/"),
        ("http:\\/\\/example.com", "http://example.com"),
    ],
)
def test_string_decoder__forward_slash_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\\\", "\\"),
        ("C:\\\\Users\\\\Name", "C:\\Users\\Name"),
        ("\\\\\\\\", "\\\\"),
    ],
)
def test_string_decoder__backslash_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("\\\"", '"'),
        ("say \\\"hello\\\"", 'say "hello"'),
        ("\\\"\\\"\\\"", '"""'),
    ],
)
def test_string_decoder__quote_escape(stream: str, expected_string: str):
    decoder = StringEscapeDecoder()
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == expected_string


def test_string_decoder__is_terminating_quote_unescaped():
    decoder = StringEscapeDecoder()
    assert decoder.is_terminating_quote('"') is True


def test_string_decoder__is_terminating_quote_escaped():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    assert decoder.is_terminating_quote('"') is False


def test_string_decoder__is_terminating_quote_other_chars():
    decoder = StringEscapeDecoder()
    assert decoder.is_terminating_quote("a") is False
    assert decoder.is_terminating_quote("'") is False
    assert decoder.is_terminating_quote(" ") is False
    assert decoder.is_terminating_quote("\\") is False


def test_string_decoder__is_terminating_quote_after_escaped_backslash():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    decoder.push("\\")
    assert decoder.is_terminating_quote('"') is True


def test_string_decoder__reset_clears_buffer():
    decoder = StringEscapeDecoder()
    decoder.push("h")
    decoder.push("e")
    decoder.push("l")
    decoder.push("l")
    decoder.push("o")
    assert decoder.buffer == "hello"
    decoder.reset()
    assert decoder.buffer == ""


def test_string_decoder__reset_clears_escape_state():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    decoder.reset()
    assert decoder.is_terminating_quote('"') is True


def test_string_decoder__buffer_property_returns_accumulated():
    decoder = StringEscapeDecoder()
    assert decoder.buffer == ""
    decoder.push("a")
    assert decoder.buffer == "a"
    decoder.push("b")
    assert decoder.buffer == "ab"
    decoder.push("c")
    assert decoder.buffer == "abc"


def test_string_decoder__push_returns_character():
    decoder = StringEscapeDecoder()
    assert decoder.push("a") == "a"
    assert decoder.push("b") == "b"


def test_string_decoder__push_returns_none_for_backslash():
    decoder = StringEscapeDecoder()
    assert decoder.push("\\") is None


def test_string_decoder__push_returns_escaped_char_after_backslash():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    assert decoder.push("n") == "\n"


def test_string_decoder__push_returns_none_for_unicode_start():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    assert decoder.push("u") is None


@pytest.mark.parametrize(
    "invalid_escape,expected_char",
    [
        ("\\x", "x"),
        ("\\a", "a"),
        ("\\z", "z"),
        ("\\1", "1"),
        ("\\@", "@"),
    ],
)
def test_string_decoder__invalid_escape_falls_back_to_literal(invalid_escape: str, expected_char: str):
    decoder = StringEscapeDecoder()
    for ch in invalid_escape:
        decoder.push(ch)
    assert decoder.buffer == expected_char


def test_string_decoder__multiple_escape_sequences():
    decoder = StringEscapeDecoder()
    stream = "line1\\nline2\\ttab\\r\\nend"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "line1\nline2\ttab\r\nend"


def test_string_decoder__escape_at_end_of_input():
    decoder = StringEscapeDecoder()
    decoder.push("a")
    decoder.push("b")
    decoder.push("\\")
    assert decoder.buffer == "ab"
    assert decoder._string_escape is True


def test_string_decoder__empty_string():
    decoder = StringEscapeDecoder()
    assert decoder.buffer == ""


def test_string_decoder__single_character():
    decoder = StringEscapeDecoder()
    decoder.push("x")
    assert decoder.buffer == "x"


def test_string_decoder__whitespace_preserved():
    decoder = StringEscapeDecoder()
    stream = "  hello  world  "
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "  hello  world  "


def test_string_decoder__unicode_characters_pass_through():
    decoder = StringEscapeDecoder()
    stream = "こんにちは"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "こんにちは"


def test_string_decoder__emoji_pass_through():
    decoder = StringEscapeDecoder()
    stream = "🎉🚀💡"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "🎉🚀💡"


def test_string_decoder__mixed_content():
    decoder = StringEscapeDecoder()
    stream = "Hello\\nWorld 🌍\\t日本語"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "Hello\nWorld 🌍\t日本語"


def test_string_decoder__very_long_string():
    decoder = StringEscapeDecoder()
    stream = "a" * 10000
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "a" * 10000


def test_string_decoder__all_escape_sequences_together():
    decoder = StringEscapeDecoder()
    stream = "\\\"\\\\\\b\\f\\n\\r\\t\\/"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == '"\\\b\f\n\r\t/'


def test_string_decoder__reset_multiple_times():
    decoder = StringEscapeDecoder()
    decoder.push("a")
    decoder.reset()
    decoder.push("b")
    decoder.reset()
    decoder.push("c")
    assert decoder.buffer == "c"


def test_string_decoder__push_after_reset():
    decoder = StringEscapeDecoder()
    decoder.push("f")
    decoder.push("i")
    decoder.push("r")
    decoder.push("s")
    decoder.push("t")
    decoder.reset()
    decoder.push("s")
    decoder.push("e")
    decoder.push("c")
    decoder.push("o")
    decoder.push("n")
    decoder.push("d")
    assert decoder.buffer == "second"


def test_string_decoder__consecutive_backslashes_odd():
    decoder = StringEscapeDecoder()
    stream = "\\\\\\\\"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "\\\\"


def test_string_decoder__consecutive_backslashes_even():
    decoder = StringEscapeDecoder()
    stream = "\\\\\\\\\\\\"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "\\\\\\"


def test_string_decoder__escape_map_access():
    decoder = StringEscapeDecoder()
    assert decoder.escape_map['"'] == '"'
    assert decoder.escape_map["\\"] == "\\"
    assert decoder.escape_map["/"] == "/"
    assert decoder.escape_map["b"] == "\b"
    assert decoder.escape_map["f"] == "\f"
    assert decoder.escape_map["n"] == "\n"
    assert decoder.escape_map["r"] == "\r"
    assert decoder.escape_map["t"] == "\t"


def test_string_decoder__unicode_escape_sets_parsing_flag():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    decoder.push("u")
    assert decoder._is_parsing_unicode is True


def test_string_decoder__unicode_escape_after_regular_text():
    decoder = StringEscapeDecoder()
    for ch in "hello\\u0041world":
        decoder.push(ch)
    assert decoder.buffer == "helloAworld"


def test_string_decoder__escaped_backslash_followed_by_quote():
    decoder = StringEscapeDecoder()
    for ch in "\\\\\\\"":
        decoder.push(ch)
    assert decoder.buffer == '\\"'


def test_string_decoder__terminating_quote_after_regular_text():
    decoder = StringEscapeDecoder()
    for ch in "hello":
        decoder.push(ch)
    assert decoder.is_terminating_quote('"') is True


def test_string_decoder__reset_mid_escape():
    decoder = StringEscapeDecoder()
    decoder.push("\\")
    decoder.push("u")
    decoder.push("0")
    decoder.push("0")
    decoder.reset()
    assert decoder.buffer == ""
    assert decoder._is_parsing_unicode is False
    assert decoder._string_escape is False


def test_string_decoder__special_json_characters():
    decoder = StringEscapeDecoder()
    stream = "{}[],:123"
    for ch in stream:
        decoder.push(ch)
    assert decoder.buffer == "{}[],:123"



import asyncio
import json
import os
from asyncio import gather, wait_for
from enum import Enum
from types import NoneType
from typing import List, Optional, Type

import pytest

from jmux.awaitable import AwaitableValue, StreamableValues
from jmux.demux import JMux
from jmux.error import (
    NotAllObjectPropertiesSetError,
)


class AsyncStreamGenerator:
    stream: str

    def __init__(self, stream):
        self.stream = stream

    async def __anext__(self):
        raise Exception("This method is not implemented")

    async def __aiter__(self):
        for char in self.stream:
            yield char
            await asyncio.sleep(0)


LOG_EMITS = os.environ.get("LOG_EMITS", "0") == "1"


def log_emit(message: str):
    if LOG_EMITS:
        print(message)


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"city_name":"Paris","country":"France"}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: i",
                "[producer] sending: t",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: n",
                "[producer] sending: a",
                "[producer] sending: m",
                "[producer] sending: e",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: P",
                "[city] received: P",
                "[producer] sending: a",
                "[city] received: a",
                "[producer] sending: r",
                "[city] received: r",
                "[producer] sending: i",
                "[city] received: i",
                "[producer] sending: s",
                "[city] received: s",
                '[producer] sending: "',
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: o",
                "[producer] sending: u",
                "[producer] sending: n",
                "[producer] sending: t",
                "[producer] sending: r",
                "[producer] sending: y",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: F",
                "[country] received: F",
                "[producer] sending: r",
                "[country] received: r",
                "[producer] sending: a",
                "[country] received: a",
                "[producer] sending: n",
                "[country] received: n",
                "[producer] sending: c",
                "[country] received: c",
                "[producer] sending: e",
                "[country] received: e",
                '[producer] sending: "',
                "[producer] sending: }",
            ],
        ),
        (
            '{"city_name": "Paris", "country": "France"}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: i",
                "[producer] sending: t",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: n",
                "[producer] sending: a",
                "[producer] sending: m",
                "[producer] sending: e",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending:  ",
                '[producer] sending: "',
                "[producer] sending: P",
                "[city] received: P",
                "[producer] sending: a",
                "[city] received: a",
                "[producer] sending: r",
                "[city] received: r",
                "[producer] sending: i",
                "[city] received: i",
                "[producer] sending: s",
                "[city] received: s",
                '[producer] sending: "',
                "[producer] sending: ,",
                "[producer] sending:  ",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: o",
                "[producer] sending: u",
                "[producer] sending: n",
                "[producer] sending: t",
                "[producer] sending: r",
                "[producer] sending: y",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending:  ",
                '[producer] sending: "',
                "[producer] sending: F",
                "[country] received: F",
                "[producer] sending: r",
                "[country] received: r",
                "[producer] sending: a",
                "[country] received: a",
                "[producer] sending: n",
                "[country] received: n",
                "[producer] sending: c",
                "[country] received: c",
                "[producer] sending: e",
                "[country] received: e",
                '[producer] sending: "',
                "[producer] sending: }",
            ],
        ),
        (
            '{\n\t"city_name": "Paris",\n\t"country": "France"\n}',
            [
                "[producer] sending: {",
                "[producer] sending: \n",
                "[producer] sending: \t",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: i",
                "[producer] sending: t",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: n",
                "[producer] sending: a",
                "[producer] sending: m",
                "[producer] sending: e",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending:  ",
                '[producer] sending: "',
                "[producer] sending: P",
                "[city] received: P",
                "[producer] sending: a",
                "[city] received: a",
                "[producer] sending: r",
                "[city] received: r",
                "[producer] sending: i",
                "[city] received: i",
                "[producer] sending: s",
                "[city] received: s",
                '[producer] sending: "',
                "[producer] sending: ,",
                "[producer] sending: \n",
                "[producer] sending: \t",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: o",
                "[producer] sending: u",
                "[producer] sending: n",
                "[producer] sending: t",
                "[producer] sending: r",
                "[producer] sending: y",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending:  ",
                '[producer] sending: "',
                "[producer] sending: F",
                "[country] received: F",
                "[producer] sending: r",
                "[country] received: r",
                "[producer] sending: a",
                "[country] received: a",
                "[producer] sending: n",
                "[country] received: n",
                "[producer] sending: c",
                "[country] received: c",
                "[producer] sending: e",
                "[country] received: e",
                '[producer] sending: "',
                "[producer] sending: \n",
                "[producer] sending: }",
            ],
        ),
    ],
)
@pytest.mark.anyio
async def test_json_demux__simple_json(stream: str, expected_operations: List[str]):
    class SCityName(JMux):
        city_name: StreamableValues[str]
        country: StreamableValues[str]

    llm_stream = AsyncStreamGenerator(stream)
    s_city = SCityName()

    city_name = ""
    country = ""
    operation_list = []

    async def consume_city():
        nonlocal city_name
        async for ch in s_city.city_name:
            op = f"[city] received: {ch}"
            log_emit(op)
            operation_list.append(op)
            city_name += ch

    async def consume_country():
        nonlocal country
        async for ch in s_city.country:
            op = f"[country] received: {ch}"
            log_emit(op)
            operation_list.append(op)
            country += ch

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            log_emit(op)
            operation_list.append(op)
            await s_city.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_city(),
        consume_country(),
    )

    parsed_json = json.loads(stream)

    assert parsed_json["city_name"] == city_name
    assert parsed_json["country"] == country

    assert operation_list == expected_operations


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"emojis":"😀😃😄😁😆😅😂🤣😊😇🙂🙃😉😌😍😘🥰😗😙😚"}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: e",
                "[producer] sending: m",
                "[producer] sending: o",
                "[producer] sending: j",
                "[producer] sending: i",
                "[producer] sending: s",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: 😀",
                "[emojis] received: 😀",
                "[producer] sending: 😃",
                "[emojis] received: 😃",
                "[producer] sending: 😄",
                "[emojis] received: 😄",
                "[producer] sending: 😁",
                "[emojis] received: 😁",
                "[producer] sending: 😆",
                "[emojis] received: 😆",
                "[producer] sending: 😅",
                "[emojis] received: 😅",
                "[producer] sending: 😂",
                "[emojis] received: 😂",
                "[producer] sending: 🤣",
                "[emojis] received: 🤣",
                "[producer] sending: 😊",
                "[emojis] received: 😊",
                "[producer] sending: 😇",
                "[emojis] received: 😇",
                "[producer] sending: 🙂",
                "[emojis] received: 🙂",
                "[producer] sending: 🙃",
                "[emojis] received: 🙃",
                "[producer] sending: 😉",
                "[emojis] received: 😉",
                "[producer] sending: 😌",
                "[emojis] received: 😌",
                "[producer] sending: 😍",
                "[emojis] received: 😍",
                "[producer] sending: 😘",
                "[emojis] received: 😘",
                "[producer] sending: 🥰",
                "[emojis] received: 🥰",
                "[producer] sending: 😗",
                "[emojis] received: 😗",
                "[producer] sending: 😙",
                "[emojis] received: 😙",
                "[producer] sending: 😚",
                "[emojis] received: 😚",
                '[producer] sending: "',
                "[producer] sending: }",
            ],
        )
    ],
)
@pytest.mark.anyio
async def test_json_demux__utf8(stream: str, expected_operations: List[str]):
    class SEmojis(JMux):
        emojis: StreamableValues[str]

    llm_stream = AsyncStreamGenerator(stream)
    s_emoji = SEmojis()

    emojis = ""
    operation_list = []

    async def consume_emojis():
        nonlocal emojis
        async for ch in s_emoji.emojis:
            op = f"[emojis] received: {ch}"
            operation_list.append(op)
            emojis += ch

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            operation_list.append(op)
            await s_emoji.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_emojis(),
    )

    parsed_json = json.loads(stream)

    assert emojis == parsed_json["emojis"]
    assert operation_list == expected_operations


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"my_str":"foo","my_int":42,"my_float":3.14,"my_bool":true,"my_enum":"value1","my_none":null}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: s",
                "[producer] sending: t",
                "[producer] sending: r",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: f",
                "[producer] sending: o",
                "[producer] sending: o",
                '[producer] sending: "',
                "[str] received: foo",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: i",
                "[producer] sending: n",
                "[producer] sending: t",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: 4",
                "[producer] sending: 2",
                "[producer] sending: ,",
                "[int] received: 42",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: f",
                "[producer] sending: l",
                "[producer] sending: o",
                "[producer] sending: a",
                "[producer] sending: t",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: 3",
                "[producer] sending: .",
                "[producer] sending: 1",
                "[producer] sending: 4",
                "[producer] sending: ,",
                "[float] received: 3.14",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: b",
                "[producer] sending: o",
                "[producer] sending: o",
                "[producer] sending: l",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: t",
                "[producer] sending: r",
                "[producer] sending: u",
                "[producer] sending: e",
                "[producer] sending: ,",
                "[bool] received: True",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: e",
                "[producer] sending: n",
                "[producer] sending: u",
                "[producer] sending: m",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: v",
                "[producer] sending: a",
                "[producer] sending: l",
                "[producer] sending: u",
                "[producer] sending: e",
                "[producer] sending: 1",
                '[producer] sending: "',
                "[enum] received: SEnum.VALUE1",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: m",
                "[producer] sending: y",
                "[producer] sending: _",
                "[producer] sending: n",
                "[producer] sending: o",
                "[producer] sending: n",
                "[producer] sending: e",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: n",
                "[producer] sending: u",
                "[producer] sending: l",
                "[producer] sending: l",
                "[producer] sending: }",
                "[none] received: None",
            ],
        )
    ],
)
@pytest.mark.anyio
async def test_json_demux__primitives(stream: str, expected_operations: List[str]):
    class SPrimitives(JMux):
        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        my_str: AwaitableValue[str]
        my_int: AwaitableValue[int]
        my_float: AwaitableValue[float]
        my_bool: AwaitableValue[bool]
        my_enum: AwaitableValue[SEnum]
        my_none: AwaitableValue[NoneType]

    llm_stream = AsyncStreamGenerator(stream)
    s_primitives = SPrimitives()

    my_str: Optional[str] = None
    my_int: Optional[int] = None
    my_float: Optional[float] = None
    my_bool: Optional[bool] = None
    my_enum: Optional[SPrimitives.SEnum] = None
    my_none: Optional[NoneType] = None
    operation_list = []

    async def consume_str():
        nonlocal my_str
        my_str = await s_primitives.my_str
        op = f"[str] received: {my_str}"
        log_emit(op)
        operation_list.append(op)

    async def consume_int():
        nonlocal my_int
        my_int = await s_primitives.my_int
        op = f"[int] received: {my_int}"
        log_emit(op)
        operation_list.append(op)

    async def consume_float():
        nonlocal my_float
        my_float = await s_primitives.my_float
        op = f"[float] received: {my_float}"
        log_emit(op)
        operation_list.append(op)

    async def consume_bool():
        nonlocal my_bool
        my_bool = await s_primitives.my_bool
        op = f"[bool] received: {my_bool}"
        log_emit(op)
        operation_list.append(op)

    async def consume_enum():
        nonlocal my_enum
        my_enum = await s_primitives.my_enum
        op = f"[enum] received: {my_enum}"
        log_emit(op)
        operation_list.append(op)

    async def consume_none():
        nonlocal my_none
        my_none = await s_primitives.my_none
        op = f"[none] received: {my_none}"
        log_emit(op)
        operation_list.append(op)

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            log_emit(op)
            operation_list.append(op)
            await s_primitives.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_str(),
        consume_int(),
        consume_float(),
        consume_bool(),
        consume_enum(),
        consume_none(),
    )

    parsed_json = json.loads(stream)

    assert my_str == parsed_json["my_str"]
    assert my_int == parsed_json["my_int"]
    assert my_float == parsed_json["my_float"]
    assert my_bool == parsed_json["my_bool"]
    assert my_enum is not None and my_enum.value == parsed_json["my_enum"]
    assert my_none == parsed_json["my_none"]
    assert operation_list == expected_operations


class SPrimitivesPartial1(JMux):
    my_str: AwaitableValue[str]
    my_int: AwaitableValue[int | NoneType]
    my_float: AwaitableValue[float]
    my_bool: AwaitableValue[bool | NoneType]
    my_none: AwaitableValue[NoneType]


@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    [
        (
            '{"key_str":"foo","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"bar","key_nested":{"key_str":"nested_value"}}',
            None,
        ),
        (
            '{"key_str":"foo","key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"bar","key_nested":{"key_str":"nested_value"}}',
            None,
        ),
        (
            '{"key_str":"foo","key_float":3.14,"key_none":null,"key_stream":"bar","key_nested":{"key_str":"nested_value"}}',
            None,
        ),
        (
            '{"key_str":"foo","key_float":3.14,"key_stream":"bar","key_nested":{"key_str":"nested_value"}}',
            None,
        ),
        (
            '{"key_str":"foo","key_float":3.14,"key_stream":"bar","key_nested":null}',
            None,
        ),
        (
            '{"key_str":"foo","key_float":3.14,"key_stream":"bar"}',
            None,
        ),
        (
            '{"key_float":3.14}',
            NotAllObjectPropertiesSetError,
        ),
        (
            '{"key_str":"foo"}',
            NotAllObjectPropertiesSetError,
        ),
    ],
)
@pytest.mark.anyio
async def test_json_demux__primitives__partial_streams(
    stream: str,
    MaybeExpectedError: Type[Exception] | None,
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int | NoneType]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool | NoneType]
        key_none: AwaitableValue[NoneType]
        key_stream: StreamableValues[str]
        key_nested: AwaitableValue[SNested | NoneType]

        arr_str: StreamableValues[str]
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_none: StreamableValues[NoneType]
        arr_nested: StreamableValues[SNested]

    llm_stream = AsyncStreamGenerator(stream)
    s_primitives = SObject()

    my_str: Optional[str] = None
    my_int: Optional[int] = None
    my_float: Optional[float] = None
    my_bool: Optional[bool] = None
    my_none: Optional[NoneType] = None
    operation_list = []

    async def consume_str():
        nonlocal my_str
        my_str = await s_primitives.key_str
        op = f"[str] received: {my_str}"
        log_emit(op)
        operation_list.append(op)

    async def consume_int():
        nonlocal my_int
        my_int = await s_primitives.key_int
        op = f"[int] received: {my_int}"
        log_emit(op)
        operation_list.append(op)

    async def consume_float():
        nonlocal my_float
        my_float = await s_primitives.key_float
        op = f"[float] received: {my_float}"
        log_emit(op)
        operation_list.append(op)

    async def consume_bool():
        nonlocal my_bool
        my_bool = await s_primitives.key_bool
        op = f"[bool] received: {my_bool}"
        log_emit(op)
        operation_list.append(op)

    async def consume_none():
        nonlocal my_none
        my_none = await s_primitives.key_none
        op = f"[none] received: {my_none}"
        log_emit(op)
        operation_list.append(op)

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            log_emit(op)
            operation_list.append(op)
            await s_primitives.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            await wait_for(
                fut=gather(
                    produce(),
                    consume_str(),
                    consume_int(),
                    consume_float(),
                    consume_bool(),
                    consume_none(),
                ),
                timeout=5,
            )
    else:
        await wait_for(
            fut=gather(
                produce(),
                consume_str(),
                consume_int(),
                consume_float(),
                consume_bool(),
                consume_none(),
            ),
            timeout=5,
        )
        parsed_json = json.loads(stream)

        assert my_str == parsed_json.get("key_str", None)
        assert my_int == parsed_json.get("key_int", None)
        assert my_float == parsed_json.get("key_float", None)
        assert my_bool == parsed_json.get("key_bool", None)
        assert my_none == parsed_json.get("key_none", None)


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"nested":{"child":"child_value"}}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: n",
                "[producer] sending: e",
                "[producer] sending: s",
                "[producer] sending: t",
                "[producer] sending: e",
                "[producer] sending: d",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: {",
                "[nested] received: <SNestedChild>",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: h",
                "[producer] sending: i",
                "[producer] sending: l",
                "[producer] sending: d",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: h",
                "[producer] sending: i",
                "[producer] sending: l",
                "[producer] sending: d",
                "[producer] sending: _",
                "[producer] sending: v",
                "[producer] sending: a",
                "[producer] sending: l",
                "[producer] sending: u",
                "[producer] sending: e",
                '[producer] sending: "',
                "[child] received: child_value",
                "[producer] sending: }",
                "[producer] sending: }",
            ],
        )
    ],
)
@pytest.mark.anyio
async def test_json_demux__nested(stream: str, expected_operations: List[str]):
    class SParent(JMux):
        class SNested(JMux):
            child: AwaitableValue[str]

        nested: AwaitableValue[SNested]

    llm_stream = AsyncStreamGenerator(stream)
    s_parent = SParent()

    nested: Optional[SParent.SNested] = None
    child: Optional[str] = None
    operation_list = []

    async def consume_nested():
        nonlocal nested, child
        nested = await s_parent.nested
        op = "[nested] received: <SNestedChild>"
        log_emit(op)
        operation_list.append(op)

        child = await nested.child
        op = f"[child] received: {child}"
        log_emit(op)
        operation_list.append(op)

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            log_emit(op)
            operation_list.append(op)
            await s_parent.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_nested(),
    )

    parsed_json = json.loads(stream)
    assert nested is not None
    assert isinstance(nested, SParent.SNested)
    assert child == parsed_json["nested"]["child"]
    assert operation_list == expected_operations


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"arr":[{"key":"value1"},{"key":"value2"}]}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: r",
                "[producer] sending: r",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: [",
                "[producer] sending: {",
                "[nested] received: <SArrayElement>",
                '[producer] sending: "',
                "[producer] sending: k",
                "[producer] sending: e",
                "[producer] sending: y",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: v",
                "[producer] sending: a",
                "[producer] sending: l",
                "[producer] sending: u",
                "[producer] sending: e",
                "[producer] sending: 1",
                '[producer] sending: "',
                "[child] received: value1",
                "[producer] sending: }",
                "[producer] sending: ,",
                "[producer] sending: {",
                "[nested] received: <SArrayElement>",
                '[producer] sending: "',
                "[producer] sending: k",
                "[producer] sending: e",
                "[producer] sending: y",
                '[producer] sending: "',
                "[producer] sending: :",
                '[producer] sending: "',
                "[producer] sending: v",
                "[producer] sending: a",
                "[producer] sending: l",
                "[producer] sending: u",
                "[producer] sending: e",
                "[producer] sending: 2",
                '[producer] sending: "',
                "[child] received: value2",
                "[producer] sending: }",
                "[producer] sending: ]",
                "[producer] sending: }",
            ],
        )
    ],
)
@pytest.mark.anyio
async def test_json_demux__object_with_array_of_objects(
    stream: str, expected_operations: List[str]
):
    class SParent(JMux):
        class SArrayElement(JMux):
            key: AwaitableValue[str]

        arr: StreamableValues[SArrayElement]

    llm_stream = AsyncStreamGenerator(stream)
    s_parent = SParent()

    arr: List[SParent.SArrayElement] = []
    child: List[str] = []
    operation_list = []

    async def consume_nested():
        nonlocal arr, child
        async for element in s_parent.arr:
            arr.append(element)
            op = "[nested] received: <SArrayElement>"
            log_emit(op)
            operation_list.append(op)

            key_value = await element.key
            await asyncio.sleep(0)
            child.append(key_value)
            op = f"[child] received: {key_value}"
            log_emit(op)
            operation_list.append(op)

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            log_emit(op)
            operation_list.append(op)
            await s_parent.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_nested(),
    )

    parsed_json = json.loads(stream)
    assert arr is not None
    assert len(arr) == 2
    assert isinstance(arr[0], SParent.SArrayElement)
    assert child[0] == parsed_json["arr"][0]["key"]
    assert operation_list == expected_operations


@pytest.mark.parametrize(
    "stream,expected_operations",
    [
        (
            '{"arr_int":[1,2],"arr_float":[1.1,2.2],"arr_bool":[true,false],"arr_str":["ab","cd"]}',
            [
                "[producer] sending: {",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: r",
                "[producer] sending: r",
                "[producer] sending: _",
                "[producer] sending: i",
                "[producer] sending: n",
                "[producer] sending: t",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: [",
                "[producer] sending: 1",
                "[producer] sending: ,",
                "[arr_int] received: 1",
                "[producer] sending: 2",
                "[producer] sending: ]",
                "[arr_int] received: 2",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: r",
                "[producer] sending: r",
                "[producer] sending: _",
                "[producer] sending: f",
                "[producer] sending: l",
                "[producer] sending: o",
                "[producer] sending: a",
                "[producer] sending: t",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: [",
                "[producer] sending: 1",
                "[producer] sending: .",
                "[producer] sending: 1",
                "[producer] sending: ,",
                "[arr_float] received: 1.1",
                "[producer] sending: 2",
                "[producer] sending: .",
                "[producer] sending: 2",
                "[producer] sending: ]",
                "[arr_float] received: 2.2",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: r",
                "[producer] sending: r",
                "[producer] sending: _",
                "[producer] sending: b",
                "[producer] sending: o",
                "[producer] sending: o",
                "[producer] sending: l",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: [",
                "[producer] sending: t",
                "[producer] sending: r",
                "[producer] sending: u",
                "[producer] sending: e",
                "[producer] sending: ,",
                "[arr_bool] received: True",
                "[producer] sending: f",
                "[producer] sending: a",
                "[producer] sending: l",
                "[producer] sending: s",
                "[producer] sending: e",
                "[producer] sending: ]",
                "[arr_bool] received: False",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: r",
                "[producer] sending: r",
                "[producer] sending: _",
                "[producer] sending: s",
                "[producer] sending: t",
                "[producer] sending: r",
                '[producer] sending: "',
                "[producer] sending: :",
                "[producer] sending: [",
                '[producer] sending: "',
                "[producer] sending: a",
                "[producer] sending: b",
                '[producer] sending: "',
                "[arr_str] received: ab",
                "[producer] sending: ,",
                '[producer] sending: "',
                "[producer] sending: c",
                "[producer] sending: d",
                '[producer] sending: "',
                "[arr_str] received: cd",
                "[producer] sending: ]",
                "[producer] sending: }",
            ],
        )
    ],
)
@pytest.mark.anyio
async def test_json_demux__object_with_array_of_primitives(
    stream: str, expected_operations: List[str]
):
    class SParent(JMux):
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_str: StreamableValues[str]

    llm_stream = AsyncStreamGenerator(stream)
    s_parent = SParent()

    arr_int: List[int] = []
    arr_float: List[float] = []
    arr_bool: List[bool] = []
    arr_str: List[str] = []
    operation_list = []

    async def consume_int_arr():
        nonlocal arr_int
        async for element in s_parent.arr_int:
            arr_int.append(element)
            op = f"[arr_int] received: {element}"
            print(op)
            operation_list.append(op)

    async def consume_float_arr():
        nonlocal arr_float
        async for element in s_parent.arr_float:
            arr_float.append(element)
            op = f"[arr_float] received: {element}"
            print(op)
            operation_list.append(op)

    async def consume_bool_arr():
        nonlocal arr_bool
        async for element in s_parent.arr_bool:
            arr_bool.append(element)
            op = f"[arr_bool] received: {element}"
            print(op)
            operation_list.append(op)

    async def consume_str_arr():
        nonlocal arr_str
        async for element in s_parent.arr_str:
            arr_str.append(element)
            op = f"[arr_str] received: {element}"
            print(op)
            operation_list.append(op)

    async def produce():
        async for ch in llm_stream:
            op = f"[producer] sending: {ch}"
            print(op)
            operation_list.append(op)
            await s_parent.feed_char(ch)
            # Yield control to allow other tasks to run
            # Necessary in the tests only, for API calls this is not needed
            await asyncio.sleep(0)

    await gather(
        produce(),
        consume_int_arr(),
        consume_float_arr(),
        consume_bool_arr(),
        consume_str_arr(),
    )

    parsed_json = json.loads(stream)
    assert len(arr_int) == 2
    assert arr_int == parsed_json["arr_int"]
    assert len(arr_float) == 2
    assert arr_float == parsed_json["arr_float"]
    assert len(arr_bool) == 2
    assert arr_bool == parsed_json["arr_bool"]
    assert len(arr_str) == 2
    assert arr_str == parsed_json["arr_str"]
    assert operation_list == expected_operations


@pytest.mark.anyio
async def test_get_current_awaitable_after_feed():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": "value"}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.key.get_current() == "value"


@pytest.mark.anyio
async def test_get_current_streamable_during_feed():
    class SObject(JMux):
        arr: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr": [1, 2, 3]}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.arr.get_current() == 3


@pytest.mark.anyio
async def test_get_current_int_array():
    class SObject(JMux):
        nums: StreamableValues[int]

    s_object = SObject()
    stream = '{"nums": [10, 20, 30]}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.nums.get_current() == 30


@pytest.mark.anyio
async def test_awaitable_closed_after_complete_parse():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": "value"}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.key._is_closed is True


@pytest.mark.anyio
async def test_streamable_closed_after_complete_parse():
    class SObject(JMux):
        arr: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr": [1, 2, 3]}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.arr._closed is True


@pytest.mark.anyio
async def test_optional_awaitable_closed_with_null():
    class SObject(JMux):
        key: AwaitableValue[str | None]

    s_object = SObject()
    stream = '{"key": null}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object.key._is_closed is True
    assert await s_object.key is None


@pytest.mark.anyio
async def test_empty_array():
    class SObject(JMux):
        arr: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr": []}'

    for ch in stream:
        await s_object.feed_char(ch)

    items = []
    async for item in s_object.arr:
        items.append(item)

    assert items == []


@pytest.mark.anyio
async def test_empty_string():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": ""}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.key == ""


@pytest.mark.anyio
async def test_empty_nested_object():
    class SNested(JMux):
        pass

    class SObject(JMux):
        nested: AwaitableValue[SNested]

    s_object = SObject()
    stream = '{"nested": {}}'

    for ch in stream:
        await s_object.feed_char(ch)

    nested = await s_object.nested
    assert isinstance(nested, SNested)


@pytest.mark.anyio
async def test_array_of_empty_strings():
    class SObject(JMux):
        arr: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr": ["", "", ""]}'

    for ch in stream:
        await s_object.feed_char(ch)

    items = []
    async for item in s_object.arr:
        items.append(item)

    assert items == ["", "", ""]


@pytest.mark.anyio
async def test_multiple_empty_arrays():
    class SObject(JMux):
        arr1: StreamableValues[str]
        arr2: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr1": [], "arr2": []}'

    for ch in stream:
        await s_object.feed_char(ch)

    items1 = []
    async for item in s_object.arr1:
        items1.append(item)

    items2 = []
    async for item in s_object.arr2:
        items2.append(item)

    assert items1 == []
    assert items2 == []


@pytest.mark.anyio
async def test_feed_chunks_simple():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    await s_object.feed_chunks('{"key": "value"}')

    assert await s_object.key == "value"


@pytest.mark.anyio
async def test_feed_chunks_partial():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    await s_object.feed_chunks('{"key": ')
    await s_object.feed_chunks('"value"}')

    assert await s_object.key == "value"


@pytest.mark.anyio
async def test_feed_chunks_with_whitespace_before_close():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    await s_object.feed_chunks('{"key": "value"   }')

    assert await s_object.key == "value"


@pytest.mark.anyio
async def test_feed_chunks_with_trailing_whitespace():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    await s_object.feed_chunks('{"key": "value"}\n\t ')

    assert await s_object.key == "value"


@pytest.mark.anyio
async def test_negative_numbers():
    class SObject(JMux):
        num: AwaitableValue[int]

    s_object = SObject()
    stream = '{"num": -42}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.num == -42


@pytest.mark.anyio
async def test_negative_float():
    class SObject(JMux):
        num: AwaitableValue[float]

    s_object = SObject()
    stream = '{"num": -3.14}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.num == -3.14


@pytest.mark.anyio
async def test_zero_values():
    class SObject(JMux):
        int_zero: AwaitableValue[int]
        float_zero: AwaitableValue[float]

    s_object = SObject()
    stream = '{"int_zero": 0, "float_zero": 0.0}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.int_zero == 0
    assert await s_object.float_zero == 0.0


@pytest.mark.anyio
async def test_scientific_notation():
    class SObject(JMux):
        num: AwaitableValue[float]

    s_object = SObject()
    stream = '{"num": 1.5e10}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.num == 1.5e10


@pytest.mark.anyio
async def test_large_integer():
    class SObject(JMux):
        num: AwaitableValue[int]

    s_object = SObject()
    stream = '{"num": 99999999999999999999}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.num == 99999999999999999999


@pytest.mark.anyio
async def test_escaped_string_content():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": "hello\\nworld\\ttab"}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.key == "hello\nworld\ttab"


@pytest.mark.anyio
async def test_string_with_quotes():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": "say \\"hello\\""}'

    for ch in stream:
        await s_object.feed_char(ch)

    assert await s_object.key == 'say "hello"'



from types import NoneType
from typing import Set, Type

import pytest

from jmux.awaitable import (
    AwaitableValue,
    IAsyncSink,
    SinkType,
    StreamableValues,
)
from jmux.demux import JMux
from jmux.error import NothingEmittedError, SinkClosedError


class NestedObject(JMux):
    key: AwaitableValue[str]


@pytest.mark.parametrize(
    "TargetType,expected_set",
    [
        (AwaitableValue[int], {int}),
        (AwaitableValue[float], {float}),
        (AwaitableValue[str], {str}),
        (AwaitableValue[bool], {bool}),
        (AwaitableValue[NestedObject], {NestedObject}),
        (AwaitableValue[int | None], {int, NoneType}),
        (AwaitableValue[float | None], {float, NoneType}),
        (AwaitableValue[str | None], {str, NoneType}),
        (AwaitableValue[bool | None], {bool, NoneType}),
        (AwaitableValue[NestedObject | None], {NestedObject, NoneType}),
        (StreamableValues[int], {int}),
        (StreamableValues[float], {float}),
        (StreamableValues[str], {str}),
        (StreamableValues[bool], {bool}),
        (StreamableValues[NestedObject], {NestedObject}),
    ],
)
def test_underlying_generic_mixin__get_underlying_generic__expected_set(
    TargetType: Type[IAsyncSink], expected_set: Set[Type]
):
    target = TargetType()
    underlying_types = target.get_underlying_generics()

    assert underlying_types == expected_set


@pytest.mark.parametrize(
    "TargetType,MaybeExpectedError",
    [
        (AwaitableValue[int], None),
        (AwaitableValue[str], None),
        (AwaitableValue[float], None),
        (AwaitableValue[bool], None),
        (AwaitableValue[NoneType], None),
        (AwaitableValue[NestedObject | None], None),
        (AwaitableValue[int | NoneType], None),
        (AwaitableValue[str | NoneType], None),
        (AwaitableValue[float | NoneType], None),
        (AwaitableValue[bool | NoneType], None),
        (AwaitableValue[bool | str | NoneType], TypeError),
        (AwaitableValue[bool | str], TypeError),
        (AwaitableValue[NestedObject | str], TypeError),
        (StreamableValues[int | None], TypeError),
        (StreamableValues[float | None], TypeError),
        (StreamableValues[str | None], TypeError),
        (StreamableValues[bool | None], TypeError),
        (StreamableValues[NestedObject | None], TypeError),
    ],
)
def test_underlying_generic_mixin__get_underlying_generic__check_instantiation(
    TargetType: Type[IAsyncSink], MaybeExpectedError: Type[Exception] | None
):
    target = TargetType()
    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            _ = target.get_underlying_generics()
    else:
        _ = target.get_underlying_generics()


@pytest.mark.anyio
async def test_awaitable_value__double_put_raises_value_error():
    av = AwaitableValue[int]()
    await av.put(42)
    with pytest.raises(ValueError, match="can only be set once"):
        await av.put(100)


@pytest.mark.anyio
async def test_awaitable_value__get_current_before_set_raises_value_error():
    av = AwaitableValue[int]()
    with pytest.raises(ValueError, match="has not been set yet"):
        av.get_current()


@pytest.mark.anyio
async def test_awaitable_value__close_without_value_non_optional_raises_nothing_emitted_error():
    av = AwaitableValue[int]()
    with pytest.raises(NothingEmittedError, match="without a value"):
        await av.close()


@pytest.mark.anyio
async def test_awaitable_value__close_without_value_optional_succeeds():
    av = AwaitableValue[int | None]()
    await av.close()
    assert av._is_closed is True


@pytest.mark.anyio
async def test_awaitable_value__double_close_raises_sink_closed_error():
    av = AwaitableValue[int]()
    await av.put(42)
    await av.close()
    with pytest.raises(SinkClosedError, match="already closed"):
        await av.close()


@pytest.mark.anyio
async def test_awaitable_value__put_after_close_raises_value_error():
    av = AwaitableValue[int]()
    await av.put(42)
    await av.close()
    with pytest.raises(ValueError, match="can only be set once"):
        await av.put(100)


@pytest.mark.anyio
async def test_awaitable_value__get_sink_type_returns_awaitable_value():
    av = AwaitableValue[int]()
    assert av.get_sink_type() == SinkType.AWAITABLE_VALUE


@pytest.mark.anyio
async def test_awaitable_value__ensure_closed_is_idempotent():
    av = AwaitableValue[int]()
    await av.put(42)
    await av.ensure_closed()
    await av.ensure_closed()
    await av.ensure_closed()
    assert av._is_closed is True


@pytest.mark.anyio
async def test_awaitable_value__ensure_closed_without_value_non_optional_raises():
    av = AwaitableValue[str]()
    with pytest.raises(NothingEmittedError):
        await av.ensure_closed()


@pytest.mark.anyio
async def test_awaitable_value__await_returns_correct_value():
    av = AwaitableValue[str]()
    await av.put("hello")
    result = await av
    assert result == "hello"


@pytest.mark.anyio
async def test_awaitable_value__await_optional_none_value():
    av = AwaitableValue[str | None]()
    await av.close()
    result = await av
    assert result is None


@pytest.mark.anyio
async def test_awaitable_value__get_current_after_put():
    av = AwaitableValue[float]()
    await av.put(3.14)
    assert av.get_current() == 3.14


@pytest.mark.anyio
async def test_awaitable_value__get_underlying_main_generic_single_type():
    av = AwaitableValue[int]()
    assert av.get_underlying_main_generic() is int


@pytest.mark.anyio
async def test_awaitable_value__get_underlying_main_generic_optional():
    av = AwaitableValue[str | None]()
    assert av.get_underlying_main_generic() is str


@pytest.mark.parametrize(
    "value",
    [
        0,
        -1,
        999999999,
        -999999999,
    ],
)
@pytest.mark.anyio
async def test_awaitable_value__various_int_values(value: int):
    av = AwaitableValue[int]()
    await av.put(value)
    assert av.get_current() == value
    assert await av == value


@pytest.mark.parametrize(
    "value",
    [
        "",
        "hello",
        "hello world",
        "unicode: こんにちは",
        "emoji: 🎉",
        "a" * 10000,
    ],
)
@pytest.mark.anyio
async def test_awaitable_value__various_string_values(value: str):
    av = AwaitableValue[str]()
    await av.put(value)
    assert av.get_current() == value
    assert await av == value


@pytest.mark.anyio
async def test_awaitable_value__bool_true():
    av = AwaitableValue[bool]()
    await av.put(True)
    assert av.get_current() is True


@pytest.mark.anyio
async def test_awaitable_value__bool_false():
    av = AwaitableValue[bool]()
    await av.put(False)
    assert av.get_current() is False


@pytest.mark.anyio
async def test_awaitable_value__nested_object():
    av = AwaitableValue[NestedObject]()
    nested = NestedObject()
    await av.put(nested)
    assert av.get_current() is nested


@pytest.mark.anyio
async def test_streamable_values__put_after_close_raises_value_error():
    sv = StreamableValues[int]()
    await sv.close()
    with pytest.raises(ValueError, match="closed sink"):
        await sv.put(42)


@pytest.mark.anyio
async def test_streamable_values__get_current_before_items_raises_value_error():
    sv = StreamableValues[int]()
    with pytest.raises(ValueError, match="not received any items"):
        sv.get_current()


@pytest.mark.anyio
async def test_streamable_values__double_close_raises_sink_closed_error():
    sv = StreamableValues[int]()
    await sv.close()
    with pytest.raises(SinkClosedError, match="already closed"):
        await sv.close()


@pytest.mark.anyio
async def test_streamable_values__get_sink_type_returns_streamable_values():
    sv = StreamableValues[int]()
    assert sv.get_sink_type() == SinkType.STREAMABLE_VALUES


@pytest.mark.anyio
async def test_streamable_values__ensure_closed_is_idempotent():
    sv = StreamableValues[str]()
    await sv.put("a")
    await sv.ensure_closed()
    await sv.ensure_closed()
    await sv.ensure_closed()
    assert sv._closed is True


@pytest.mark.anyio
async def test_streamable_values__aiter_yields_all_items():
    sv = StreamableValues[int]()
    await sv.put(1)
    await sv.put(2)
    await sv.put(3)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == [1, 2, 3]


@pytest.mark.anyio
async def test_streamable_values__empty_stream_iteration():
    sv = StreamableValues[int]()
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == []


@pytest.mark.anyio
async def test_streamable_values__large_number_of_items():
    sv = StreamableValues[int]()
    count = 1000
    for i in range(count):
        await sv.put(i)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == list(range(count))


@pytest.mark.anyio
async def test_streamable_values__get_current_returns_last_item():
    sv = StreamableValues[str]()
    await sv.put("first")
    assert sv.get_current() == "first"
    await sv.put("second")
    assert sv.get_current() == "second"
    await sv.put("third")
    assert sv.get_current() == "third"


@pytest.mark.anyio
async def test_streamable_values__get_underlying_main_generic():
    sv = StreamableValues[float]()
    assert sv.get_underlying_main_generic() is float


@pytest.mark.parametrize(
    "values",
    [
        [1],
        [1, 2],
        [1, 2, 3, 4, 5],
        list(range(100)),
    ],
)
@pytest.mark.anyio
async def test_streamable_values__various_list_lengths(values: list):
    sv = StreamableValues[int]()
    for v in values:
        await sv.put(v)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == values


@pytest.mark.parametrize(
    "value",
    [
        "",
        "hello",
        "unicode: 日本語",
        "emoji: 🚀🎉",
    ],
)
@pytest.mark.anyio
async def test_streamable_values__various_string_values(value: str):
    sv = StreamableValues[str]()
    await sv.put(value)
    assert sv.get_current() == value


@pytest.mark.anyio
async def test_streamable_values__float_values():
    sv = StreamableValues[float]()
    await sv.put(1.5)
    await sv.put(2.5)
    await sv.put(3.5)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == [1.5, 2.5, 3.5]


@pytest.mark.anyio
async def test_streamable_values__bool_values():
    sv = StreamableValues[bool]()
    await sv.put(True)
    await sv.put(False)
    await sv.put(True)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == [True, False, True]


@pytest.mark.anyio
async def test_streamable_values__nested_objects():
    sv = StreamableValues[NestedObject]()
    obj1 = NestedObject()
    obj2 = NestedObject()
    await sv.put(obj1)
    await sv.put(obj2)
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == [obj1, obj2]


@pytest.mark.anyio
async def test_streamable_values__get_current_after_close():
    sv = StreamableValues[int]()
    await sv.put(42)
    await sv.close()
    assert sv.get_current() == 42


@pytest.mark.anyio
async def test_awaitable_value__close_after_put_succeeds():
    av = AwaitableValue[int]()
    await av.put(42)
    await av.close()
    result = await av
    assert result == 42


@pytest.mark.anyio
async def test_awaitable_value__get_underlying_main_generic_nested_object():
    av = AwaitableValue[NestedObject]()
    assert av.get_underlying_main_generic() is NestedObject


@pytest.mark.anyio
async def test_streamable_values__aiter_string_values():
    sv = StreamableValues[str]()
    await sv.put("hello")
    await sv.put("world")
    await sv.close()

    items = []
    async for item in sv:
        items.append(item)

    assert items == ["hello", "world"]


@pytest.mark.anyio
async def test_streamable_values__get_underlying_main_generic_string_type():
    sv = StreamableValues[str]()
    assert sv.get_underlying_main_generic() is str
